/*

This software is open source pursuant to the "BSD License."

   Copyright (c) 2003, Josh Goldfoot
   All rights reserved.
   
   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are
   met:
   
     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in
       the documentation and/or other materials provided with the
       distribution.
     * Neither the name of Goldfoot Biochemical nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.
       
   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
   HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
   TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
   USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
   DAMAGE.

*/

/* Code version 2.0 */
#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "StarterRsc.h"
#include <SysEvtMgr.h>
#include <ExgMgr.h>
#include "FreeCurrencyRsc.h"
#include "StarterRsc.h"


/***********************************************************************
 *
 *	Entry Points
 *
 ***********************************************************************/


/***********************************************************************
 *
 *	Internal Structures
 *
 ***********************************************************************/
 enum  field_type { Unit1Qty, Unit2Qty, Unit1Put, Unit2Put, Unit1Name, Unit2Name, 
 			blank };
 			
struct myPreferenceType
	{
	Long		prefversion;
	char		pUnit1Name[80];
	char		pUnit2Name[80];
	double		pUnit1Qty, pUnit2Qty, pUnit1Put, pUnit2Put;
	UInt		current_record_index;
	} ;
	
typedef struct 
	{
	UInt8 replaceme;
	} StarterAppInfoType;

typedef StarterAppInfoType* StarterAppInfoPtr;

class	cRecord
{
public:
	cRecord();
	cRecord(const CharPtr unit1name, const CharPtr unit2name, double unit1qty, double unit2qty);
	~cRecord();
	Byte		version;
	char		unit_1_name[80];
	char		unit_2_name[80];
	double		unit_1_qty;
	double		unit_2_qty;
};


// Constants

const unsigned long appFileCreator = 'fCUR';
const unsigned short appVersionNum = 0x01;
const unsigned short appPrefID = 0x00;
const unsigned short appPrefVersionNum = 0x01;
const unsigned long version30 = 0x03000000;
const	CharPtr		currency_DB_name = "CurrencyDatabase-fCUR";
const	ULong		currency_DB_type = 'crdb';

// Globals

myPreferenceType 			prefs;
DmOpenRef					gDB;
CharPtr						popup_label;


// Function prototypes

void			print_names();
double			ascii2double(CharPtr s);
Short			double2ascii(const double x, CharPtr s);
Short			do_conversion(field_type	which_changed);
void			record2fields();
Boolean			handle_keydown_event(EventPtr eventP);
FieldPtr		set_field_text_from_handle(Word field_id, VoidHand text_handle);
FieldPtr		set_field_text_from_string(Word field_id, CharPtr string_pointer);
void			clear_field_text(Word field_id);
CharPtr			get_field_text(Word field_id);
Boolean			handle_button_event(EventPtr eventP);
void			swap_fields(Word field1, Word field2);
void			flip_values();
void			populate_new_database();
UInt			add_record_to_database(const CharPtr unit1name, const CharPtr unit2name, double unit1qty, 
			double unit2qty, UInt actual_index);
void			get_record_name(UInt item_num, CharPtr list_entry);
void			currency_list_draw_function(Int16 item_num, RectanglePtr bounds, CharPtr* pUnused);
Boolean			select_new_currency(Int16 selection);
void			do_save_command();
void			do_new_command();
void			do_delete_command();
void			do_restore_to_original_command();
void			create_new_database();
void			receive_currency(ExgSocketPtr socket_pointer, Boolean app_active);
void			goto_item(GoToParamsPtr goToParams, Boolean launchingApp);
void 			search_find(FindParamsPtr findParams);
void			react_to_field_change(Word field_id);
void			save_preferences();
void			fill_in_fields();
void			send_currency(const int beam_only);

float cvt_atof(char *strP,short *fracP);
void cvt_ultoa(unsigned long value, char *strP);
unsigned long cvt_atoul(char *strP);
void format_number(float value,short frac,char *strP);

// Define the minimum OS version we support (2.0 for now).
#define kOurMinVersion	sysMakeROMVersion(2,0,0,sysROMStageRelease,0)
#define kPalmOS10Version	sysMakeROMVersion(1,0,0,sysROMStageRelease,0)


/***********************************************************************
 *
 *	Internal Functions
 *
 ***********************************************************************/


/***********************************************************************
 *
 * FUNCTION:    RomVersionCompatible
 *
 * DESCRIPTION: This routine checks that a ROM version is meet your
 *              minimum requirement.
 *
 * PARAMETERS:  requiredVersion - minimum rom version required
 *                                (see sysFtrNumROMVersion in SystemMgr.h 
 *                                for format)
 *              launchFlags     - flags that indicate if the application 
 *                                UI is initialized.
 *
 * RETURNED:    error code or zero if rom is compatible
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
	UInt32 romVersion;

	// See if we're on in minimum required version of the ROM or later.
	FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
	if (romVersion < requiredVersion)
		{
		if ((launchFlags & (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
			(sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
			{
			FrmAlert (RomIncompatibleAlert);
		
			// Palm OS 1.0 will continuously relaunch this app unless we switch to 
			// another safe one.
			if (romVersion <= kPalmOS10Version)
				{
				AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch, NULL);
				}
			}
		
		return sysErrRomIncompatible;
		}

	return errNone;
}


/***********************************************************************
 *
 * FUNCTION:    GetObjectPtr
 *
 * DESCRIPTION: This routine returns a pointer to an object in the current
 *              form.
 *
 * PARAMETERS:  formId - id of the form to display
 *
 * RETURNED:    void *
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void * GetObjectPtr(UInt16 objectID)
{
	FormPtr frmP;

	frmP = FrmGetActiveForm();
	return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));
}


/***********************************************************************
 *
 * FUNCTION:    MainFormInit
 *
 * DESCRIPTION: This routine initializes the MainForm form.
 *
 * PARAMETERS:  frm - pointer to the MainForm form.
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void MainFormInit(FormPtr frmP)
{

	
	ListPtr		currency_list_ptr = (ListPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainCurrenciesList));

	LstSetListChoices( currency_list_ptr, NULL, DmNumRecords (gDB) );
	LstSetDrawFunction(currency_list_ptr, currency_list_draw_function);
	LstSetSelection(currency_list_ptr, 0);
	
	
}

void	fill_in_fields()
{
	CharPtr		buf = (CharPtr) MemPtrNew(128);
	
	if (prefs.current_record_index == -1)
		select_new_currency(0);
	else {
		set_field_text_from_string(MainUnit1NameField, prefs.pUnit1Name);
		set_field_text_from_string(MainUnit2NameField, prefs.pUnit2Name);
		double2ascii(prefs.pUnit1Qty, buf);
		set_field_text_from_string(MainUnit1QtyField, buf);
		double2ascii(prefs.pUnit2Qty, buf);
		set_field_text_from_string(MainUnit2QtyField, buf);
		double2ascii(prefs.pUnit1Put, buf);
		set_field_text_from_string(MainUnit1PutField, buf);
		double2ascii(prefs.pUnit2Put, buf);
		set_field_text_from_string(MainUnit2PutField, buf);
		
		get_record_name(prefs.current_record_index, popup_label);
		CtlSetLabel (ControlPtr(GetObjectPtr(MainCurrenciesPopTrigger)), popup_label);

	}
	
	MemPtrFree( (VoidPtr) buf);
	print_names();

}

/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:  command  - menu item id
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean MainFormDoCommand(UInt16 command)
{
	Boolean handled = false;
	FieldPtr	active_field = NULL;
	Word		active_field_id = 0;
	FormPtr		frmP = FrmGetActiveForm();
	
	// If we are in the main form, check to see if there is an active field.
	if (frmP) {
		UInt16 focus = FrmGetFocus(frmP);
		if (focus != noFocus) {
			active_field_id = FrmGetObjectId(frmP, focus);	
			if (active_field_id == noFocus)
				active_field = NULL;
			else
				active_field = (FieldPtr) GetObjectPtr(active_field_id);
		}
	}
	
	switch (command)
		{
		case MainOptionsAboutFreeCurrency:
			FrmAlert(AboutAlert);
			handled = true;
			break;
		
		case EditUndo:
			handled = true;
			if (active_field)
				FldUndo(active_field);
			break;
		
		case EditGraffitiHelp:
			SysGraffitiReferenceDialog(referenceDefault);
			handled = true;
			break;
		
		case EditKeyboard:
			SysKeyboardDialog (kbdDefault);
			react_to_field_change(active_field_id);
			handled = true;
			break;
		
		case EditCopy:
			if (active_field)
				FldCopy(active_field);
			break;
		
		case EditCut:
			if (active_field) {
				FldCut(active_field);
				react_to_field_change(active_field_id);
			}
			break;
		
		case EditPaste:
			if (active_field){
				FldPaste(active_field);
				react_to_field_change(active_field_id);
			}

			break;
		
		case CurrencyNew:
			do_new_command();
			handled = true;
			break;
		
		case CurrencySave:
			do_save_command();
			handled = true;
			break;
		
		case CurrencyDelete:
			do_delete_command();
			handled = true;
			break;
		
		case CurrencyBeam:
			send_currency(true);
			handled = true;
			break;
		
		case CurrencySend:
			send_currency(false);
			handled = true;
			break;
		
		case CurrencyRestoretooriginal:
			do_restore_to_original_command();
			handled = true;
			break;

		}
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "MainForm" of this application.
 *
 * PARAMETERS:  eventP  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean MainFormHandleEvent(EventPtr eventP)
{
	Boolean handled = false;
	FormPtr frmP = NULL;

	switch (eventP->eType) 
		{
		case menuEvent:
			handled = MainFormDoCommand(eventP->data.menu.itemID);
			return handled;
			break;

		case frmOpenEvent:
			frmP = FrmGetActiveForm();
			MainFormInit( frmP);
			FrmDrawForm ( frmP);
			fill_in_fields();
			handled = true;
			break;
			
		case frmUpdateEvent:
			// To do any custom drawing here, first call FrmDrawForm(), then do your
			// drawing, and then set handled to true.
			break;
		case keyDownEvent:
			handled = handle_keydown_event(eventP);
			break;
		
		case popSelectEvent:
			handled = false;
			if (eventP->data.popSelect.controlID == MainCurrenciesPopTrigger) 
				handled = select_new_currency(eventP->data.popSelect.selection);
			break;
			
		
		case ctlSelectEvent:
			handled = handle_button_event(eventP);
			break;

		case frmGotoEvent:
			prefs.current_record_index = eventP->data.frmGoto.recordNum;
			select_new_currency(prefs.current_record_index);
			if (eventP->data.frmGoto.matchFieldNum) {
				FieldPtr selectedField = FieldPtr(GetObjectPtr(eventP->data.frmGoto.matchFieldNum));
				FldSetScrollPosition(selectedField, eventP->data.frmGoto.matchPos);
				FrmSetFocus(frmP, FrmGetObjectIndex(frmP, eventP->data.frmGoto.matchFieldNum));
				FldSetSelection(selectedField, eventP->data.frmGoto.matchPos, eventP->data.frmGoto.matchPos + eventP->data.frmGoto.matchLen);

			}
			handled = true;
			break;

		default:
			break;
		
		}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 *              handler for the form loaded.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean AppHandleEvent(EventPtr eventP)
{
	UInt16 formId;
	FormPtr frmP;

	if (eventP->eType == frmLoadEvent)
		{
		// Load the form resource.
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);

		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time is receives an
		// event.
		switch (formId)
			{
			case MainForm:
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;

			default:
//				ErrFatalDisplay("Invalid Form Load Event");
				break;

			}
		return true;
		}
	
	return false;
}


/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void AppEventLoop(void)
{
	UInt16 error;
	EventType event;

	do {
		EvtGetEvent(&event, evtWaitForever);

		if (! SysHandleEvent(&event))
			if (! MenuHandleEvent(0, &event, &error))
				if (! AppHandleEvent(&event))
					FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:     AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     Err value 0 if nothing went wrong
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err AppStart(void)
{
    Word prefsSize;

	// Read the saved preferences / saved-state information.
	prefsSize = sizeof(myPreferenceType);
	if (PrefGetAppPreferences(appFileCreator, appPrefID, &prefs, &prefsSize, true) == 
		noPreferenceFound)
		{
			MemSet( VoidPtr(&prefs), prefsSize, 0);
			prefs.prefversion = 1;
			StrCopy(prefs.pUnit1Name, "xNew");
			StrCopy(prefs.pUnit2Name, "xNew");
			prefs.pUnit1Qty = 1;
			prefs.pUnit2Qty = 1;
			prefs.pUnit1Put = 0;
			prefs.pUnit2Put = 0;
			prefs.current_record_index = -1;
			
		}
	
	gDB = DmOpenDatabaseByTypeCreator(currency_DB_type, appFileCreator, dmModeReadWrite);
	if (! gDB) {
		create_new_database();
	}

	popup_label = (CharPtr) MemPtrNew(168);
	
	
   return 0;
}


/***********************************************************************
 *
 * FUNCTION:    AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void AppStop(void)
{
	save_preferences();
   
    MemPtrFree( (VoidPtr) popup_label);
	
	if (gDB != NULL)
		DmCloseDatabase(gDB);
	
	FrmCloseAllForms();
}


/***********************************************************************
 *
 * FUNCTION:    StarterPalmMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 *
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static UInt32 FreeCurrencyPilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
	Err error;
	Boolean	app_active;


	error = RomVersionCompatible (version30, launchFlags);
	if (error) return (error);


	switch (cmd)
		{
		case sysAppLaunchCmdNormalLaunch:
			error = AppStart();
			if (error) 
				return error;
				
			FrmGotoForm(MainForm);
			AppEventLoop();
			AppStop();
			break;
			
		case sysAppLaunchCmdSyncNotify:
			// Use this if I have more than one data type, but I don't...
			break;
		
		
		case sysAppLaunchCmdExgReceiveData:
			app_active = (launchFlags & sysAppLaunchFlagSubCall);
			receive_currency(ExgSocketPtr(cmdPBP), app_active);
			break;
		
		case sysAppLaunchCmdSaveData:
			// Do nothing; we don't need to save data
			break;
		
		case sysAppLaunchCmdGoTo:
			Boolean		launched = launchFlags & sysAppLaunchFlagNewGlobals;
			if (launched) {
				error = AppStart();
				if (! error) {
					goto_item((GoToParamsPtr) cmdPBP, launched);
					AppEventLoop();
					AppStop();
				}
			} else
				goto_item((GoToParamsPtr) cmdPBP, launched);
			break;
		
		case sysAppLaunchCmdFind:
			search_find((FindParamsPtr)cmdPBP);

		default:
			break;

		}
	
	return 0;
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
UInt32 PilotMain( UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	return FreeCurrencyPilotMain(cmd, cmdPBP, launchFlags);
}





// print_names:  Gets the names of the currencies, and prints them
// in the lower part of the form.
void			print_names()
{
	CharPtr			cpUnit1Name = get_field_text(MainUnit1NameField);
	CharPtr			cpUnit2Name = get_field_text(MainUnit2NameField);
	//RectangleType	r;
	//RctSetRectangle(&r, 104, 75, 56, 50);
	//WinEraseRectangle(&r, 0);
	
	FormType *formP = FrmGetActiveForm();
	UInt16 Unit1Index, Unit2Index;
	
	Unit1Index = FrmGetObjectIndex (formP, MainUnit1NameLabel);
	Unit2Index = FrmGetObjectIndex (formP, MainUnit2NameLabel);
	
	FrmHideObject (formP, Unit1Index);
	FrmCopyLabel (formP, MainUnit1NameLabel, cpUnit1Name);
	FrmShowObject (formP, Unit1Index);
	FrmHideObject (formP, Unit2Index);
	FrmCopyLabel (formP, MainUnit2NameLabel, cpUnit2Name);
	FrmShowObject (formP, Unit2Index);
	
	//WinDrawChars(cpUnit1Name, StrLen(cpUnit1Name), 104, 80);
	//WinDrawChars(cpUnit2Name, StrLen(cpUnit2Name), 104, 112);
}

// ascii2double: Given a string containing a number (s), return a double
// representing the value of the string.
double			ascii2double(CharPtr s)
{	
	if (s != NULL) {
		if (StrLen(s) > 0)
			return cvt_atof(s, NULL);
		else
			return 0;
	} else
		return 0;
}

// double2ascii: given a double and a pointer to a string buffer,
// put an ASCII representation of the double into the buffer, using normal
// (i.e. non-scientific) notation, rounding to hundredths.

Short			double2ascii(const double x, CharPtr s)
{

	format_number(x, 2, s);
	return StrLen(s);

}

// do_conversion: the main currency conversion routine.  "which_changed"
// tells the routine which of the fields the user has changed.  Based on that,
// the routine decides which field to recalculate, and does so.
Short			do_conversion(field_type	which_changed)
{
	// Get the text of all the fields.
	CharPtr			cpUnit1Qty = get_field_text(MainUnit1QtyField);
	CharPtr			cpUnit2Qty = get_field_text(MainUnit2QtyField);
	if ((StrLen(cpUnit1Qty) == 0) || (StrLen(cpUnit2Qty) == 0))
		return 0;
		
	double			dUnit1Qty, dUnit1Put, dUnit2Qty, dUnit2Put;
	
	// Convert from strings to floating point numbers for calculating.
	dUnit1Qty = ascii2double(cpUnit1Qty);
	dUnit2Qty = ascii2double(cpUnit2Qty);
	dUnit1Put = ascii2double(get_field_text(MainUnit1PutField));
	dUnit2Put = ascii2double(get_field_text(MainUnit2PutField));
	
	char		buf[255];
	
	// If either of the main unit quantities are zero, quit to avoid division
	// by zero errors.
	if ((dUnit1Qty == 0) || (dUnit2Qty == 0)) {
		clear_field_text(MainUnit1PutField);
		clear_field_text(MainUnit2PutField);
		return 0;
	}
	
	// Do the calculation.  Note that if the "quantity" fields change, only
	// unit 2 put gets changed.
	
	if (which_changed != Unit2Put) {
		dUnit2Put = dUnit1Put * dUnit2Qty / dUnit1Qty;
		double2ascii(dUnit2Put, buf);
		set_field_text_from_string(MainUnit2PutField, buf);
	}

	if (which_changed != Unit1Put) {
		dUnit1Put = dUnit2Put * dUnit1Qty / dUnit2Qty;
		double2ascii(dUnit1Put, buf); 
		set_field_text_from_string(MainUnit1PutField, buf);
	}	
	

	return 0;
}



// handle_keydown_event: called after the user enters a character.
Boolean			handle_keydown_event(EventPtr eventP)
{
	FormPtr		frmP = FrmGetActiveForm();
	UInt new_index, index_max;

	if (! frmP)
		return false;
	if (eventP->data.keyDown.modifiers & commandKeyMask) {
		new_index = prefs.current_record_index;
		index_max = DmNumRecords (gDB) - 1;
		switch (eventP->data.keyDown.chr) {
			case pageUpChr:
				if (new_index == 0)
					new_index = index_max;
				else
					new_index--;
				select_new_currency(new_index);	
				break;
			case pageDownChr:
				if (new_index == index_max)
					new_index = 0;
				else
					new_index++;
				select_new_currency(new_index);	
				break;
		}
	} else {
		UInt16 focus = FrmGetFocus(frmP);
		
		Word		active_field_id = FrmGetObjectId(frmP, FrmGetFocus(FrmGetActiveForm()));
		FldHandleEvent ( FieldPtr(FrmGetObjectPtr(frmP, 
							FrmGetObjectIndex(frmP, active_field_id))), eventP);
		react_to_field_change( active_field_id );
	}	
	return true;
}



FieldPtr		set_field_text_from_string(Word field_id, CharPtr string_pointer)
{
	FieldPtr field = (FieldPtr) GetObjectPtr(field_id);

	MemHandle      h;
	short handle_size = StrLen(string_pointer) + 1;

	h = FldGetTextHandle(field);
	if (h) {
		Err	err;

		FldSetTextHandle(field, NULL);
		
		err = MemHandleResize(h, handle_size );
		if (err) {
			FldSetTextHandle(field, h); 
			return field;
		}
	} else {
		h = MemHandleNew(handle_size);
		if (!h)
			return field;
	}

	StrNCopy((Char *) MemHandleLock(h), string_pointer, handle_size);
	MemHandleUnlock(h);

	FldSetTextHandle(field, h);
	FldDrawField(field);
	return field;
} 


void			clear_field_text(Word field_id)
{
	set_field_text_from_string(field_id, "");
}

CharPtr			get_field_text(Word field_id)
{
	FormPtr			frmP = FrmGetActiveForm();
	return FldGetTextPtr(FieldPtr(FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, field_id))));
}

Boolean			handle_button_event(EventPtr eventP)
{
	Boolean		handled = false;
	
	switch (eventP->data.ctlSelect.controlID) {
		case MainSaveButton:
			handled = true;
			do_save_command();
			break;
		case MainNewButton:
			handled = true;
			do_new_command();
			break;
		case MainDeleteButton:
			handled = true;
			do_delete_command();
			break;
		default:
			handled = false;
			break;
		
	}

	return handled;
}

void			do_save_command()
{
	char*	question = (char*) MemPtrNew(256);
	if (! question)
		return;
	char*	record_name = (char*) MemPtrNew(164);
	if (! record_name)
		return;
	get_record_name(prefs.current_record_index, record_name);
	StrPrintF(question, "Really save over \"%s?\"", record_name);
	
	Word	button_tapped = FrmCustomAlert (QueryAlert, question, NULL, NULL);
	MemPtrFree(record_name);
	MemPtrFree(question);
	if (button_tapped == QueryNo)
		return;

	cRecord			my_record(get_field_text(MainUnit1NameField), 
				get_field_text(MainUnit2NameField), 
				ascii2double(get_field_text(MainUnit1QtyField)), 
				ascii2double(get_field_text(MainUnit2QtyField)) );
	
	VoidHand	record_handle = DmGetRecord( gDB, prefs.current_record_index );
	if (! record_handle)
		return;

	VoidPtr		record_pointer = MemHandleLock( record_handle );
	UInt		the_size = sizeof(cRecord);
	DmWrite( record_pointer, 0, &my_record, the_size );
	MemHandleUnlock( record_handle );
	DmReleaseRecord( gDB, prefs.current_record_index, true);
	get_record_name(prefs.current_record_index, popup_label);
	CtlSetLabel (ControlPtr(GetObjectPtr(MainCurrenciesPopTrigger)), popup_label);

	
}

void			do_new_command()
{
	UInt	new_index = add_record_to_database("USA Dollar", "Foreign", 1, 3, prefs.current_record_index);
	LstSetListChoices( ListPtr(GetObjectPtr(MainCurrenciesList)), 
		NULL, DmNumRecords (gDB) );
	select_new_currency(new_index);
}

void			do_delete_command()
{
	if (DmNumRecords(gDB) == 1) {
		FrmCustomAlert(InfoAlert, "Sorry, that's your last record.", NULL, NULL);
		return;
	}
	
	char*	question = (char*) MemPtrNew(256);
	if (! question)
		return;
	char*	record_name = (char*) MemPtrNew(164);
	if (! record_name)
		return;
	get_record_name(prefs.current_record_index, record_name);
	StrPrintF(question, "Really delete \"%s?\"", record_name);
	
	Word	button_tapped = FrmCustomAlert (QueryAlert, question, NULL, NULL);
	MemPtrFree(record_name);
	MemPtrFree(question);
	if (button_tapped == QueryNo)
		return;
	
	Err		myErr = DmRemoveRecord (gDB, prefs.current_record_index);
	if (prefs.current_record_index > 0) {
		prefs.current_record_index--;
	}
	LstSetListChoices( ListPtr(GetObjectPtr(MainCurrenciesList)), 
		NULL, DmNumRecords (gDB) );
	select_new_currency(prefs.current_record_index);

}

void			do_restore_to_original_command()
{
	
	Word	button_tapped = FrmCustomAlert (QueryAlert, "Really delete ALL records and replace them with a default set?", NULL, NULL);
	if (button_tapped == QueryNo)
		return;
	
	UInt	num_records = DmNumRecords(gDB);
	while (num_records > 0) {
		DmRemoveRecord (gDB, 0);
		num_records--;
	}
	populate_new_database();
	LstSetListChoices( ListPtr(GetObjectPtr(MainCurrenciesList)), 
		NULL, DmNumRecords (gDB) );
	select_new_currency(0);
	save_preferences();
}


cRecord::cRecord()
{
	version = 1;
	StrCopy(unit_1_name, "");
	StrCopy(unit_2_name, "");
	unit_1_qty = 1;
	unit_2_qty = 1;
}

cRecord::cRecord(const CharPtr unit1name, const CharPtr unit2name, double unit1qty, double unit2qty)
{
	version = 1;
	StrNCopy(unit_1_name, unit1name, 79);
	StrNCopy(unit_2_name, unit2name, 79);
	unit_1_qty = unit1qty;
	unit_2_qty = unit2qty;
}

cRecord::~cRecord()
{

}


UInt			add_record_to_database(const CharPtr unit1name, const CharPtr unit2name, double unit1qty, 
			double unit2qty, UInt actual_index)
{
	cRecord			my_record(unit1name, unit2name, unit1qty, unit2qty);
	ULong			record_size = sizeof(cRecord);
	VoidHand		record_handle = DmNewHandle( gDB, record_size );
	if (! record_handle)
		return -1;
	VoidPtr			record_pointer = MemHandleLock( record_handle );
	DmWrite( record_pointer, 0, &my_record, record_size );
	MemHandleUnlock( record_handle );
	DmAttachRecord( gDB, &actual_index,  record_handle, NULL );
	return	actual_index;
}

void			populate_new_database()
{
	// Defaults to major USA and German trading partners
	// Country abbreviations are taken from ISO 3166 for consistency
	// Rates current as of January 19, 2003
	
	add_record_to_database("USA Dollar", "AUS Dollar", 1, 1.69 , dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "BRA Real", 1, 3.57  , dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "CAN Dollar", 1, 1.54 , dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "CHE Franc", 1, 1.37, dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "CHN Yuan", 1, 8.29, dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "Euro", 1.07, 1 , dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "GBR Pound", 1.61, 1, dmMaxRecordIndex);   //
	add_record_to_database("USA Dollar", "IND Rupee", 1, 47.99, dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "JPN Yen", 1, 117.83, dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "KOR Won", 1, 1173.50 , dmMaxRecordIndex);  //
	add_record_to_database("USA Dollar", "MEX Peso", 1, 10.59, dmMaxRecordIndex); //
	add_record_to_database("USA Dollar", "TWN Dollar", 1, 34.53, dmMaxRecordIndex); //
	add_record_to_database("Euro", "CHE Franc", 1, 1.46, dmMaxRecordIndex); //
	add_record_to_database("Euro", "GBR Pound", 1.51, 1, dmMaxRecordIndex); //
	add_record_to_database("Euro", "JPN Yen", 1, 125.75, dmMaxRecordIndex); //
}

// Given a database index (zero-based), return in list_entry the "name" of that
// record, e.g. "USA Dollar/FRA Franc"
void			get_record_name(UInt item_num, CharPtr list_entry)
{
	if (list_entry == NULL)
		return;
	
	if (! gDB) {
		StrCopy(list_entry, "Error!");
		return;
	}
	
	VoidHand 	record_handle = DmQueryRecord (gDB, item_num);
	if (! record_handle)
		return;
	VoidPtr		record_pointer = MemHandleLock( record_handle );	
	cRecord*	the_record = (cRecord*) record_pointer;
	StrPrintF(list_entry, "%s/%s", the_record->unit_1_name, the_record->unit_2_name);

	MemHandleUnlock(record_handle);
}


// The draw function called by the OS to draw each entry in the list function.
// The best way I know of to implement a dynamic list.
void			currency_list_draw_function(Int16 item_num, RectanglePtr bounds, CharPtr* pUnused)
{
	FormPtr		pForm = FrmGetActiveForm();
	CharPtr		list_entry = (CharPtr) MemPtrNew(164);
	if (! list_entry)
		return;

	get_record_name(item_num, list_entry);

	WinDrawChars(list_entry, StrLen(list_entry), bounds->topLeft.x, bounds->topLeft.y);
	MemPtrFree( (VoidPtr) list_entry);
}

// select_new_currency: called when the user picks a new currency, or when the
// app otherwise has to move to a new currency.  Updates the displays.
Boolean			select_new_currency(Int16 selection)
{
	if (! gDB)
		return false;
	
	VoidHand 	record_handle = DmQueryRecord (gDB, selection);
	if (! record_handle)
		return true;
	VoidPtr		record_pointer = MemHandleLock( record_handle );	
	cRecord*	the_record = (cRecord*) record_pointer;
	
	CharPtr		buf = (CharPtr) MemPtrNew(128);
	
	set_field_text_from_string(MainUnit1NameField, the_record->unit_1_name);
	set_field_text_from_string(MainUnit2NameField, the_record->unit_2_name);
	double2ascii(the_record->unit_1_qty, buf);
	set_field_text_from_string(MainUnit1QtyField, buf);
	double2ascii(the_record->unit_2_qty, buf);
	set_field_text_from_string(MainUnit2QtyField, buf);
	StrCopy(buf, "0");
	set_field_text_from_string(MainUnit1PutField, buf);
	set_field_text_from_string(MainUnit2PutField, buf);

	MemHandleUnlock(record_handle);
	
	print_names();
	prefs.current_record_index = selection;
	get_record_name(selection, popup_label);
	FormPtr frmP = FrmGetActiveForm();
	ListPtr		currency_list_ptr = (ListPtr) FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, MainCurrenciesList));
	if (currency_list_ptr)
		LstSetSelection (currency_list_ptr, selection);
	CtlSetLabel (ControlPtr(GetObjectPtr(MainCurrenciesPopTrigger)), popup_label);
	
	MemPtrFree( (VoidPtr) buf);
	
	return true;
}

void			create_new_database()
{
	Err	err = DmCreateDatabase(0, currency_DB_name, appFileCreator, currency_DB_type, false);
	if (err) return;
	gDB = DmOpenDatabaseByTypeCreator(currency_DB_type, appFileCreator, dmModeReadWrite);
	if (! gDB) return;
	populate_new_database();
}

void		send_currency(const int beam_only)
{
	Err 			err;
	ExgSocketType   exgSocket;
	VoidHand	    record_handle;
	ULong			record_size;
	VoidPtr			record_pointer;
	
	// Fill out exgSocket with the description of the data
	MemSet(&exgSocket, sizeof(exgSocket), 0);
	exgSocket.description = (CharPtr) MemPtrNew(128);
	if (! exgSocket.description)
		return;
	get_record_name(prefs.current_record_index, exgSocket.description);
	if (beam_only)
		exgSocket.name = "?_beam:FreeCurrency";
	else
		exgSocket.name = "?_send;_beam:FreeCurrency";
	exgSocket.target = appFileCreator;
	
	// Load the record
	record_handle = DmQueryRecord( gDB, prefs.current_record_index );
	if (! record_handle)
		return;
	record_pointer = MemHandleLock( record_handle );
	record_size = MemPtrSize(record_pointer);

	// Tell Exchange Manager to send the data
	err = ExgPut(&exgSocket);
	if (err)
		return;
	ExgSend(&exgSocket, record_pointer, record_size, &err);
	err = ExgDisconnect(&exgSocket, err);
	
	MemHandleUnlock( record_handle );
	MemPtrFree(exgSocket.description);
}

void		receive_currency(ExgSocketPtr socket_pointer, Boolean app_active)
{
	Err err;
	UInt32  buffer_len;
	void* data_buffer;
	UInt			record_index;
	DmOpenRef*		local_db;

	// Prepare a buffer to receive one record.
	buffer_len = sizeof(cRecord) + 1;
	data_buffer = MemPtrNew(buffer_len);
	if (! data_buffer)
		return;

	// Receive the record.
	err = ExgAccept(socket_pointer);
	buffer_len = ExgReceive(socket_pointer, data_buffer, buffer_len, &err);
	ExgDisconnect(socket_pointer, err);
	if (err || buffer_len == 0)
		return;
	
	// Obtain a reference to the database.  (It is very possible that we aren't the active app, so
	// we can't assume the global gDB is available.)
	if (app_active) 
		local_db = &gDB;
	else {
		local_db = (DmOpenRef*) MemPtrNew(sizeof(DmOpenRef));
		*local_db = DmOpenDatabaseByTypeCreator(currency_DB_type, appFileCreator, dmModeReadWrite);
		if (! *local_db)
			return;
	}

	// Add the record to our database
	VoidHand	record_handle = DmNewHandle( *local_db, buffer_len );
	if (!record_handle)
		return;
	VoidPtr		record_pointer = MemHandleLock( record_handle );
	DmWrite( record_pointer, 0, data_buffer, buffer_len );
	MemHandleUnlock( record_handle );
	record_index = 0;
	DmAttachRecord( *local_db, &record_index, record_handle, NULL );

	// Set the "goto" params so that the OS will send us a "goto" event to take us to the
	// newly-received record.
	DmRecordInfo(*local_db, record_index, NULL, &socket_pointer->goToParams.uniqueID,
		NULL);
	DmOpenDatabaseInfo(*local_db, &socket_pointer->goToParams.dbID, NULL, NULL,
		&socket_pointer->goToParams.dbCardNo, NULL);
	socket_pointer->goToParams.recordNum = record_index;
	socket_pointer->goToCreator = appFileCreator;
	
	if (! app_active) {
		DmCloseDatabase( *local_db );
		MemPtrFree( (VoidPtr) local_db);
	}
	

}


void			goto_item(GoToParamsPtr goToParams, Boolean launchingApp)
{
	EventType		event;
	UInt			recordNum = goToParams->recordNum;
	
	if (! launchingApp) {
		ULong		uniqueID;
		
		DmRecordInfo(gDB, recordNum, NULL, &uniqueID, NULL);
		DmFindRecordByID(gDB, uniqueID, &recordNum);
	}
	
	FrmGotoForm(MainForm);
	
	MemSet (&event, 0, sizeof(EventType));
	
	event.eType = frmGotoEvent;
	event.data.frmGoto.formID = MainForm;
	event.data.frmGoto.recordNum = recordNum;
	event.data.frmGoto.matchPos = goToParams->matchPos;
	event.data.frmGoto.matchLen = goToParams->searchStrLen;
	event.data.frmGoto.matchFieldNum = goToParams->matchFieldNum;
	event.data.frmGoto.matchCustom = goToParams->matchCustom;

	EvtAddEventToQueue(&event);
}


// search_find: Handles the "Find" open application message.
// searches the database for the find string, and sends self
// a goto event if it finds it.
void 			search_find(FindParamsPtr find_params)
{
	Boolean 			done, truncate;
	CharPtr				unit1_qty_str, unit2_qty_str, scratch_string;
	DmOpenRef 			local_db;
	DmSearchStateType	search_state;
	Err 				error;
	LocalID 			database_id;
	RectangleType 		r;
	SWord 				string_length, pixel_width;
	UInt 				card_number, field_number, record_number;
	VoidHand 			record_handle;
	VoidPtr				record;
	Word 				pos;

	
	find_params->more = false;	
	card_number = 0;
	error = DmGetNextDatabaseByTypeCreator(true, &search_state, currency_DB_type, appFileCreator, true,
		 &card_number, &database_id);
	if (error) 
		return;
	local_db = DmOpenDatabase(card_number, database_id, find_params->dbAccesMode);
	if (! local_db)
		return;

	// Display the heading line.
	scratch_string = (CharPtr) MemPtrNew(168);
	unit1_qty_str = (CharPtr) MemPtrNew(128);
	unit2_qty_str = (CharPtr) MemPtrNew(128);
	StrCopy(scratch_string, "Currency");
	done = FindDrawHeader(find_params, scratch_string);
	
	if (done)
		find_params->more = true;
	else {
		record_number = find_params->recordNum;	
		cRecord*		record_ptr;
		
		while (true) {
			Boolean match = false;
			
			if ( (record_number % 10 == 0) && EvtSysEventAvail(true)) {
				find_params->more = true;
				break;
			} 

			record_handle = DmQueryNextInCategory(local_db, &record_number, dmAllCategories);
			if (! record_handle)
				break;
			
			record = MemHandleLock(record_handle);
			record_ptr = (cRecord*) record;
			double2ascii(record_ptr->unit_1_qty, unit1_qty_str);
			double2ascii(record_ptr->unit_2_qty, unit2_qty_str);
			
			if ((match = FindStrInStr((CharPtr) record_ptr->unit_1_name, find_params->strToFind, &pos)) != false)
				field_number = MainUnit1NameField;
			else if ((match = FindStrInStr((CharPtr) record_ptr->unit_2_name, find_params->strToFind, &pos)) != false)
				field_number = MainUnit2NameField; 
			else if ((match = FindStrInStr(unit1_qty_str, find_params->strToFind, &pos)) != false)
				field_number = MainUnit1QtyField;
			else if ((match = FindStrInStr( unit2_qty_str, find_params->strToFind, &pos)) != false)
				field_number = MainUnit2QtyField; 
				
			if (match) {
				done = FindSaveMatch(find_params, record_number, pos, field_number, 0, 
					card_number, database_id);
				if (done) 
					break;
				StrPrintF(scratch_string, "%s/%s", record_ptr->unit_1_name, record_ptr->unit_2_name);
				FindGetLineBounds(find_params, &r);
				string_length = StrLen(scratch_string);
				pixel_width = r.extent.x;
				FntCharsInWidth(scratch_string, &pixel_width, &string_length, &truncate);
				WinDrawChars(scratch_string, string_length, r.topLeft.x, r.topLeft.y);
	
				find_params->lineNumber++;
			}
			MemHandleUnlock(record_handle);
			if (done)
				break;
			record_number++;
		}
	}	
	DmCloseDatabase(local_db);	
	MemPtrFree(scratch_string);
	MemPtrFree( (VoidPtr) unit1_qty_str);
	MemPtrFree( (VoidPtr) unit2_qty_str);

}


void			react_to_field_change(Word field_id)
{
	switch ( field_id ) {
	case MainUnit1QtyField:
		do_conversion(Unit1Qty);
		break;
	case MainUnit2QtyField:
		do_conversion(Unit2Qty);
		break;
	case MainUnit1PutField:
		do_conversion(Unit1Put);
		break;
	case MainUnit2PutField:
		do_conversion(Unit2Put);
		break;
	case MainUnit1NameField:
		print_names();
		break;
	case MainUnit2NameField:
		print_names();
		break;
	default:
		break;
	}

}

// ASCII <-> float conversion routines.  Thanks to Mike Margerum





// converts an ASCII string to an unsigned long
unsigned long cvt_atoul(char *strP)
{
unsigned long value;

    for( value = 0; (*strP >= '0') && (*strP <= '9'); )
    {
	value *= 10;
	value += *strP++ - '0';
    }

    return(value);
}

/* float cvt_atof(char *srcP, short *fracP)
 *    converts the passed string to a float. If fracP is not 0, the
 *    number of fractional digits found will be returned there.
 *    An acceptable string is of the form [+-]digits[.digits]; leading
 *    whitespace is ignored.
*/

float cvt_atof(char *strP,short *fracP)					/* convert str to fp */
{
	int i;
	int neg;
	char *cP;
	unsigned long ipart,fpart;
	float rslt;
	char	thousand_separator, decimal_separator;
	LocGetNumberSeparators ((NumberFormatType) PrefGetPreference(prefNumberFormat), 
			&thousand_separator, &decimal_separator);

    while( (*strP == ' ') || (*strP == '\t') )		/* trim whitespace */
	++strP;

    if( *strP == '-' )					/* check for neg. */
    {
	neg = 1;
	++strP;
    }
    else
    {
	neg = 0;

	if( *strP == '+' )				/* drop any sign */
	    ++strP;
    }

    ipart = cvt_atoul(strP);			/* get integer part */
    if( fracP )
	*fracP = 0;

    if( (cP = StrChr(strP,decimal_separator)) )
    {
	i = StrLen(++cP);
	fpart = cvt_atoul(cP);			/* and frac part */
	rslt = (float)fpart;

	if( fracP )
	    *fracP = i;
	while( i-- > 0 )			/* scale as needed */
	    rslt /= 10.0;

	rslt += (float)ipart;
    }
    else
	rslt = (float)ipart;

    return( neg?-rslt:rslt );
}


// converts an unsigned long into an ASCII string
void cvt_ultoa(unsigned long value, char *strP)			
{
int i;
char *cP;
char tmpstr[16];		/* a 32 bit long can give up to 10 digits */

    if( value <= 9 )			/* a quick special case */
    {
	*strP++ = value + '0';
	*strP = '\0';
	return;
    }

    /* We convert the number backwards */
    for( cP = tmpstr+9, i = 10; i-- > 0; )
    {
	*cP-- = (value % 10) + '0';
	value /= 10;
    }

    tmpstr[10] = '\0';

    for( cP = tmpstr; *cP && (*cP == '0'); )
	++cP;
    
    StrCopy(strP,cP);
}

/* void format_number(float src, short digits, char *destP)
 *    converts the floating point src to a string; the value of digits
 *    determines the number of fractional digits to produce.
 *    The result will be rounded by adding 1/2 of the least-significant
 *    digit position's scaling value to the fractional part and then truncating
 *    the fractional part; e.g. 3.6 is rounded by adding .05 to the value.
 */

void format_number(float value,short frac,char *strP)			/* format a number into str */
{
	int i;
	unsigned long ipart, fpart;
	float limit;
	char str[16];
	char	thousand_separator, decimal_separator;
	LocGetNumberSeparators ((NumberFormatType) PrefGetPreference(prefNumberFormat), 
			&thousand_separator, &decimal_separator);

    if( frac > (sizeof(str) - 1) )		/* limit frac digits */
	frac = sizeof(str) - 1;

    if( value < 0.0 )
    {
	*strP++ = '-';
	value = -value;
    }

    ipart = (long)value;
    value -= (float)ipart;			/* recover frac part as int */
    for( limit = 1.0, i = frac; i-- > 0; )
    {
	value *= 10.0;
	limit *= 10.0;
    }

    value += 0.5;				/* do some rounding */
    if( value >= limit )
    {
	fpart = 0;				/* overflowed */
	ipart++;
    }
    else
	fpart = (unsigned long)(value);

    cvt_ultoa(ipart,strP);

    if( frac )
    {
	cvt_ultoa(fpart,str);
	strP += StrLen(strP);
	*strP++ = decimal_separator;

	for( i = StrLen(str); i++ < frac; )
	    *strP++ = '0';			/* need some padding */
	StrCopy(strP,str);			/* and now the value */
    }
}

void		save_preferences()
{
	StrNCopy(prefs.pUnit1Name, get_field_text(MainUnit1NameField), 79);
	StrNCopy(prefs.pUnit2Name, get_field_text(MainUnit2NameField), 79);
	prefs.pUnit1Qty = ascii2double(get_field_text(MainUnit1QtyField));
	prefs.pUnit2Qty = ascii2double(get_field_text(MainUnit2QtyField));
	prefs.pUnit1Put = ascii2double(get_field_text(MainUnit1PutField));
	prefs.pUnit2Put = ascii2double(get_field_text(MainUnit2PutField));
   
	PrefSetAppPreferences (appFileCreator, appPrefID, appPrefVersionNum, 
		&prefs, sizeof (prefs), true);

}